using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Diagnostics;
using System.ComponentModel;

namespace Examples
{
	/// <summary>
	/// A specialized module used for comparing 4D tetrahedral meshes. It assigns a new point attribute to the evaluated mesh that contains distance to the target mesh.
	/// </summary>
	[ModuleInfo("Libor Va", "A specialized module used for comparing 4D tetrahedral meshes. It assigns a new point attribute to the evaluated mesh that contains distance to the target mesh.")]
	[PortInfo("Evaluated mesh", "A mesh whose points will be evaluated against the target mesh")]
	[PortInfo("Target mesh", "A mesh against which will be performed the evaluation")]
	[PortInfo("Resulting mesh", "A shallow copy of the evaluated mesh with added point attribute distance")]
    class AnimationDistanceEvaluator:Module
    {
		/// <summary>
		/// Default constructor, adds input and output ports
		/// </summary>
        public AnimationDistanceEvaluator()
        {
            AddInPort("Evaluated mesh", typeof(UnstrGrid));
            AddInPort("Target mesh", typeof(UnstrGrid));
            AddOutPort("Resulting mesh", typeof(UnstrGrid));
        }

        private bool paramsChanged = true;

		private int xDiv = 10, yDiv = 10, zDiv = 10, tDiv = 10;

		/// <summary>
		/// Subdivision grid resolution in the X direction
		/// </summary>
		[Description("Subdivision grid resolution in the X direction")]
		public int XDiv
		{
			get { return (xDiv); }
			set { xDiv = value; }
		}

		/// <summary>
		/// Subdivision grid resolution in the Y direction
		/// </summary>
		[Description("Subdivision grid resolution in the Y direction")]
		public int YDiv
		{
			get { return (yDiv); }
			set { yDiv = value; }
		}

		/// <summary>
		/// Subdivision grid resolution in the Z direction
		/// </summary>
		[Description("Subdivision grid resolution in the Z direction")]
		public int ZDiv
		{
			get { return (zDiv); }
			set { zDiv = value; }
		}

		/// <summary>
		/// Subdivision grid resolution in the T direction
		/// </summary>
		[Description("Subdivision grid resolution in the T direction")]
		public int TDiv
		{
			get { return (tDiv); }
			set { tDiv = value; }
		}

        TetrahedronPrecomputedData[] precomputedData;

        UnstrGrid result, target = null;

        private int[] checks;        
        private int step;
		private List<int>[, , ,] grid;
		UniformDataArray tetrahedra, points;

        private double t = 0.0;

		/// <summary>
		/// A T value taht will be used if the evaluated mesh is only 3D
		/// </summary>
        [Description("A T value taht will be used if the evaluated mesh is only 3D")]
		public double T
        {
            get
            {
                return (t);
            }
            set
            {
                t = value;
            }
        }        

        private TetrahedronPrecomputedData[] PrecomputeData(UniformDataArray tetrahedra, UniformDataArray points, out int edgecount)
        {
            Dictionary<Line, int> edges = new Dictionary<Line, int>();
            TetrahedronPrecomputedData[] result = new TetrahedronPrecomputedData[tetrahedra.Count];
            for (int i = 0; i < tetrahedra.Count; i++)
            {
                Tetrahedron t = (Tetrahedron)tetrahedra[i];
                TetrahedronPrecomputedData tdata = new TetrahedronPrecomputedData();
                tdata.v1 = (PointND)points[t.V2] - (PointND)points[t.V1];
                tdata.v2 = (PointND)points[t.V3] - (PointND)points[t.V1];
                tdata.v3 = (PointND)points[t.V4] - (PointND)points[t.V1];
                tdata.n = PointND.crossProduct(tdata.v1, tdata.v2, tdata.v3);
				tdata.nDotN = tdata.n.dotProduct(tdata.n);
                tdata.tetrahedronDet = PointToTetrahedronDistanceEvaluator.det4(tdata.v1.Values, tdata.v2.Values, tdata.v3.Values, tdata.n.Values);

                Line l;
                int index;

                if (t.V1 > t.V2) l = new Line(t.V1, t.V2);
                else l = new Line(t.V2, t.V1);
                if (edges.TryGetValue(l,out index))
                {
                    tdata.e01 = index;
                }
                else
                {
                    tdata.e01 = edges.Count;
                    edges.Add(l, tdata.e01);
                }

                if (t.V1 > t.V3) l = new Line(t.V1, t.V3);
                else l = new Line(t.V3, t.V1);
                if (edges.TryGetValue(l, out index))
                {
                    tdata.e02 = index;
                }
                else
                {
                    tdata.e02 = edges.Count;
                    edges.Add(l, tdata.e02);
                }

                if (t.V1 > t.V4) l = new Line(t.V1, t.V4);
                else l = new Line(t.V4, t.V1);
                if (edges.TryGetValue(l, out index))
                {
                    tdata.e03 = index;
                }
                else
                {
                    tdata.e03 = edges.Count;
                    edges.Add(l, tdata.e03);
                }

                if (t.V2 > t.V3) l = new Line(t.V2, t.V3);
                else l = new Line(t.V3, t.V2);
                if (edges.TryGetValue(l, out index))
                {
                    tdata.e12 = index;
                }
                else
                {
                    tdata.e12 = edges.Count;
                    edges.Add(l, tdata.e12);
                }

                if (t.V2 > t.V4) l = new Line(t.V2, t.V4);
                else l = new Line(t.V4, t.V2);
                if (edges.TryGetValue(l, out index))
                {
                    tdata.e13 = index;
                }
                else
                {
                    tdata.e13 = edges.Count;
                    edges.Add(l, tdata.e13);
                }

                if (t.V3 > t.V4) l = new Line(t.V3, t.V4);
                else l = new Line(t.V4, t.V3);
                if (edges.TryGetValue(l, out index))
                {
                    tdata.e23 = index;
                }
                else
                {
                    tdata.e23 = edges.Count;
                    edges.Add(l, tdata.e23);
                }

                result[i] = tdata;
            }
            edgecount = edges.Count;
            return (result);
        }

        private void TetrahedronBoundingBox(Tetrahedron t, UniformDataArray points, out double[] mins, out double[] maxs)
        {
            mins = new double[]{ double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity };
            maxs = new double[]{ double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity };

            PointND p = (PointND)points[t.V1];
            if (p[0] < mins[0]) mins[0] = p[0];
            if (p[1] < mins[1]) mins[1] = p[1];
            if (p[2] < mins[2]) mins[2] = p[2];
            if (p[3] < mins[3]) mins[3] = p[3];
            if (p[0] > maxs[0]) maxs[0] = p[0];
            if (p[1] > maxs[1]) maxs[1] = p[1];
            if (p[2] > maxs[2]) maxs[2] = p[2];
            if (p[3] > maxs[3]) maxs[3] = p[3];

            p = (PointND)points[t.V2];
            if (p[0] < mins[0]) mins[0] = p[0];
            if (p[1] < mins[1]) mins[1] = p[1];
            if (p[2] < mins[2]) mins[2] = p[2];
            if (p[3] < mins[3]) mins[3] = p[3];
            if (p[0] > maxs[0]) maxs[0] = p[0];
            if (p[1] > maxs[1]) maxs[1] = p[1];
            if (p[2] > maxs[2]) maxs[2] = p[2];
            if (p[3] > maxs[3]) maxs[3] = p[3];

            p = (PointND)points[t.V3];
            if (p[0] < mins[0]) mins[0] = p[0];
            if (p[1] < mins[1]) mins[1] = p[1];
            if (p[2] < mins[2]) mins[2] = p[2];
            if (p[3] < mins[3]) mins[3] = p[3];
            if (p[0] > maxs[0]) maxs[0] = p[0];
            if (p[1] > maxs[1]) maxs[1] = p[1];
            if (p[2] > maxs[2]) maxs[2] = p[2];
            if (p[3] > maxs[3]) maxs[3] = p[3];

            p = (PointND)points[t.V4];
            if (p[0] < mins[0]) mins[0] = p[0];
            if (p[1] < mins[1]) mins[1] = p[1];
            if (p[2] < mins[2]) mins[2] = p[2];
            if (p[3] < mins[3]) mins[3] = p[3];
            if (p[0] > maxs[0]) maxs[0] = p[0];
            if (p[1] > maxs[1]) maxs[1] = p[1];
            if (p[2] > maxs[2]) maxs[2] = p[2];
            if (p[3] > maxs[3]) maxs[3] = p[3];                        
        }

        private List<int>[, , ,] CreateGrid(int[] d, UnstrGrid target, UniformDataArray tetrahedra, out PointND min, out double[] steps)
        {
            // initialize result
            List<int>[, , ,] result = new List<int>[d[0], d[1], d[2], d[3]];
            for (int x = 0; x < d[0]; x++)
                for (int y = 0; y < d[1]; y++)
                    for (int z = 0; z < d[2]; z++)
                        for (int t = 0; t < d[3]; t++)
                            result[x, y, z, t] = new List<int>();

            PointND MinPoint = new PointND(4);
            PointND MaxPoint = new PointND(4);
            for (int j = 0; j < 4; j++)
            {
                MinPoint[j] = double.MaxValue;
                MaxPoint[j] = double.MinValue;
            }

            for (int i = 0; i < target.Points.Count; i++)
            {
                PointND p = (PointND)target.Points[i];
                for (int j = 0; j < 4; j++)
                {
                    if (p[j] < MinPoint[j]) MinPoint[j] = p[j];
                    if (p[j] > MaxPoint[j]) MaxPoint[j] = p[j];
                }
            }


            BoundingBox bb = target.BoundingBox;
            double[] sizes = new double[4];
            sizes[0] = (bb.MaxPoint[0] - bb.MinPoint[0])*1.01;
            sizes[1] = (bb.MaxPoint[1] - bb.MinPoint[1])*1.01;
            sizes[2] = (bb.MaxPoint[2] - bb.MinPoint[2])*1.01;
            sizes[3] = (bb.MaxPoint[3] - bb.MinPoint[3])*1.01;

            min = ((PointND)bb.MaxPoint - (PointND)bb.MinPoint) * (-0.005) + (PointND)bb.MinPoint;

            steps = new double[4];
            steps[0] = sizes[0] / d[0];
            steps[1] = sizes[1] / d[1];
            steps[2] = sizes[2] / d[2];
            steps[3] = sizes[3] / d[3];

			int sumCellCount = 0;

            for (int index = 0;index<tetrahedra.Count;index++)
            {
                Tetrahedron t = (Tetrahedron)tetrahedra[index];
                double[] mins, maxs;
                TetrahedronBoundingBox(t, target.Points, out mins, out maxs);
                
                int[] minI = {0,0,0,0};
                int[] maxI = {d[0], d[1], d[2], d[3]};

                // find X span
                double x = min[0] + steps[0];
                while (mins[0] > x)
                {
                    minI[0]++;
                    x += steps[0];
                }
                maxI[0] = minI[0];
                while (maxs[0] > x)
                {
                    maxI[0]++;
                    x += steps[0];
                }

                // find Y span
                x = min[1] + steps[1];
                while (mins[1] > x)
                {
                    minI[1]++;
                    x += steps[1];
                }
                maxI[1] = minI[1];
                while (maxs[1] > x)
                {
                    maxI[1]++;
                    x += steps[1];
                }

                // find z span
                x = min[2] + steps[2];
                while (mins[2] > x)
                {
                    minI[2]++;
                    x += steps[2];
                }
                maxI[2] = minI[2];
                while (maxs[2] > x)
                {
                    maxI[2]++;
                    x += steps[2];
                }

                // find t span
                x = min[3] + steps[3];
                while (mins[3] > x)
                {
                    minI[3]++;
                    x += steps[3];
                }
                maxI[3] = minI[3];
                while (maxs[3] > x)
                {
                    maxI[3]++;
                    x += steps[3];
                }

				int cellCount = maxI[0] - minI[0] + 1;
				for (int i = 1; i < 4; i++)
					cellCount *= (maxI[i] - minI[i] + 1);				

				if (cellCount <= 2)
				{

					// add tetrahedron to given slots
					for (int xi = minI[0]; xi <= maxI[0]; xi++)
						for (int yi = minI[1]; yi <= maxI[1]; yi++)
							for (int zi = minI[2]; zi <= maxI[2]; zi++)
								for (int ti = minI[3]; ti <= maxI[3]; ti++)
									result[xi, yi, zi, ti].Add(index);
				}
				else
				{
					cellCount = 0;
					PointND v1 = (PointND)points[t.V2] - (PointND)points[t.V1];
					PointND v2 = (PointND)points[t.V3] - (PointND)points[t.V1];
					PointND v3 = (PointND)points[t.V4] - (PointND)points[t.V1];
					PointND n = PointND.crossProduct(v1, v2, v3);
					bool[, , ,] evals = new bool[maxI[0] - minI[0] + 2, maxI[1] - minI[1] + 2, maxI[2] - minI[2] + 2, maxI[3] - minI[3] + 2];
					for (int xi = minI[0]; xi <= (maxI[0]+1); xi++)
						for (int yi = minI[1]; yi <= (maxI[1]+1); yi++)
							for (int zi = minI[2]; zi <= (maxI[2]+1); zi++)
								for (int ti = minI[3]; ti <= (maxI[3] + 1); ti++)
								{
									PointND p = min + new PointND(xi * steps[0], yi * steps[1], zi * steps[2], ti * steps[3]);
									p = p - (PointND)points[t.V1];
									evals[xi - minI[0], yi - minI[1], zi - minI[2], ti - minI[3]] = (p.dotProduct(n) > 0);
								}
					for (int xi = 0; xi <= (maxI[0]-minI[0]); xi++)
						for (int yi = 0; yi <= (maxI[1]-minI[1]); yi++)
							for (int zi = 0; zi <= (maxI[2]-minI[2]); zi++)
								for (int ti = 0; ti <= (maxI[3]-minI[3]); ti++)
								{
									bool allTrue =
										evals[xi, yi, zi, ti] &&
										evals[xi, yi, zi, ti + 1] &&
										evals[xi, yi, zi + 1, ti] &&
										evals[xi, yi, zi + 1, ti + 1] &&
										evals[xi, yi + 1, zi, ti] &&
										evals[xi, yi + 1, zi, ti + 1] &&
										evals[xi, yi + 1, zi + 1, ti] &&
										evals[xi, yi + 1, zi + 1, ti + 1] &&
										evals[xi + 1, yi, zi, ti] &&
										evals[xi + 1, yi, zi, ti + 1] &&
										evals[xi + 1, yi, zi + 1, ti] &&
										evals[xi + 1, yi, zi + 1, ti + 1] &&
										evals[xi + 1, yi + 1, zi, ti] &&
										evals[xi + 1, yi + 1, zi, ti + 1] &&
										evals[xi + 1, yi + 1, zi + 1, ti] &&
										evals[xi + 1, yi + 1, zi + 1, ti + 1];
									bool someTrue =
										evals[xi, yi, zi, ti] ||
										evals[xi, yi, zi, ti + 1] ||
										evals[xi, yi, zi + 1, ti] ||
										evals[xi, yi, zi + 1, ti + 1] ||
										evals[xi, yi + 1, zi, ti] ||
										evals[xi, yi + 1, zi, ti + 1] ||
										evals[xi, yi + 1, zi + 1, ti] ||
										evals[xi, yi + 1, zi + 1, ti + 1] ||
										evals[xi + 1, yi, zi, ti] ||
										evals[xi + 1, yi, zi, ti + 1] ||
										evals[xi + 1, yi, zi + 1, ti] ||
										evals[xi + 1, yi, zi + 1, ti + 1] ||
										evals[xi + 1, yi + 1, zi, ti] ||
										evals[xi + 1, yi + 1, zi, ti + 1] ||
										evals[xi + 1, yi + 1, zi + 1, ti] ||
										evals[xi + 1, yi + 1, zi + 1, ti + 1];

									if ((!allTrue) && (someTrue))
									{
										result[xi + minI[0], yi + minI[1], zi + minI[2], ti + minI[3]].Add(index);
										cellCount++;
									}
								}					
				}
				sumCellCount += cellCount;
            }

			Debug.WriteLine("Average cell count: " + (double)sumCellCount / tetrahedra.Count);

            return (result);
        }

        private bool cellMakesSense(int[] coords, int[] pointcoords, PointND min, double[] steps, PointND point, double current)
        {            
            double[] diff = new double[4];
            for (int i = 0; i < 4; i++)
            {
                if (coords[i] == pointcoords[i])
                {
                    diff[i] = 0;
                }
                else if (coords[i] > pointcoords[i])
                {
                    diff[i] = coords[i] * steps[i] + min[i] - point[i];
                }
                else
                {
                    diff[i] = point[i] - ((coords[i]+1) * steps[i] + min[i]);
                }
            }
            double minDistSq = diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2] + diff[3] * diff[3];
            if ((current * current) > minDistSq) return (true);
            else return (false);            
            //return (true);
        }

        private double checkCell(int[] coords, PointND point, int pi)
        {
            double min = double.PositiveInfinity;
            foreach (int i in grid[coords[0], coords[1], coords[2], coords[3]])
            {
                if (checks[i] != step)
                {
                    Tetrahedron t = (Tetrahedron)tetrahedra[i];
                    double d;
                    
                        d = PointToTetrahedronDistanceEvaluator.EvaluateDistance(point,
                        precomputedData[i],
                        pi,
                        t.V1, t.V2, t.V3, t.V4,
                        (PointND)points[t.V1],
                        (PointND)points[t.V2],
                        (PointND)points[t.V3],
                        (PointND)points[t.V4]);
                    
                    if (d < min) min = d;
                    checks[i] = step;
                }
            }
            return (min);
        }

        private bool isTetrahedronObtuse(Tetrahedron t, MveArray points)
        {
            PointND p1 = (PointND)points[t.V1];
            PointND p2 = (PointND)points[t.V2];
            PointND p3 = (PointND)points[t.V3];
            PointND p4 = (PointND)points[t.V4];

            PointND v1 = p2 - p1;
            PointND v2 = p3 - p1;
            PointND v3 = p4 - p1;
            PointND v4 = p3 - p2;
            PointND v5 = p4 - p2;
            PointND v6 = p4 - p3;

            if (v1.dotProduct(v2) < 0) return (true);
            if (v2.dotProduct(v3) < 0) return (true);
            if (v3.dotProduct(v1) < 0) return (true);

            if (v1.dotProduct(v4) > 0) return (true);
            if (v1.dotProduct(v5) > 0) return (true);
            if (v4.dotProduct(v5) < 0) return (true);

            if (v2.dotProduct(v6) > 0) return (true);
            if (v2.dotProduct(v4) < 0) return (true);
            if (v4.dotProduct(v6) > 0) return (true);

            if (v3.dotProduct(v5) < 0) return (true);
            if (v5.dotProduct(v6) < 0) return (true);
            if (v6.dotProduct(v3) < 0) return (true);

            return (false);
        }

        private double distanceTo(PointND point, int[] divs, PointND min, double[] steps, int pi)
        {
            int[,] spans = new int[4,2];

            int[] initCell = new int[4];

            int[] coords;

            // find cell where point lies
            spans[0, 0] = spans[0, 1] = initCell[0] = (int)Math.Floor((point[0] - min[0]) / steps[0]);
            spans[1, 0] = spans[1, 1] = initCell[1] = (int)Math.Floor((point[1] - min[1]) / steps[1]);
            spans[2, 0] = spans[2, 1] = initCell[2] = (int)Math.Floor((point[2] - min[2]) / steps[2]);
            spans[3, 0] = spans[3, 1] = initCell[3] = (int)Math.Floor((point[3] - min[3]) / steps[3]);
            for (int i = 0; i < 4; i++)
            {
                if (spans[i, 0] < 0) spans[i, 0] = 0;
                if (spans[i, 1] < 0) spans[i, 1] = 0;
                if (spans[i, 0] >= divs[i]) spans[i, 0] = divs[i]-1;
                if (spans[i, 1] >= divs[i]) spans[i, 1] = divs[i]-1;
            }

            double distance = double.PositiveInfinity;

            // check the cell
            double d = checkCell(new int[] { spans[0, 0], spans[1, 0], spans[2, 0], spans[3, 0] }, point, pi);
            if (d < distance) distance = d;

            bool working = true;

            while (working)
            {
                working = false;                                

                // try enlarge x
                int coord = 0;
                int nx = spans[coord, 1] + 1; // new x
                if ((nx < divs[coord]) && (distance > ((min[coord] + nx * steps[coord])-point[coord])))
                {
                    // it makes sense, check all cells with nx
                    for (int y = spans[1,0];y<=spans[1,1];y++)
                        for (int z = spans[2,0];z<=spans[2,1];z++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{nx, y, z, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 1] = nx;
                }

                // try enlarge y
                coord = 1;
                int ny = spans[coord, 1] + 1; // new y
                if ((ny < divs[coord]) && (distance > ((min[coord] + ny * steps[coord]) - point[coord])))
                {
                    // it makes sense, check all cells with ny
                    for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                        for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{x, ny, z, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 1] = ny;
                }

                // try enlarge z
                coord = 2;
                int nz = spans[coord, 1] + 1; // new z
                if ((nz < divs[coord]) && (distance > ((min[coord] + nz * steps[coord]) - point[coord])))
                {
                    // it makes sense, check all cells with nz
                    for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
                        for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{x, y, nz, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 1] = nz;
                }

                // try enlarge t
                coord = 3;
                int nt = spans[coord, 1] + 1; // new x
                if ((nt < divs[coord]) && (distance > ((min[coord] + nt * steps[coord]) - point[coord])))
                {
                    // it makes sense, check all cells with nx
                    for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
                        for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
                            for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                            {
                                coords = new int[]{x, y, z, nt};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 1] = nt;
                }

                // try shrink x
                coord = 0;
                nx = spans[coord, 0] - 1; // new x
                if ((nx >=0) && (distance > (point[coord] - (min[coord] + (nx+1) * steps[coord]))))
                {
                    // it makes sense, check all cells with nx
                    for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
                        for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{nx, y, z, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 0] = nx;
                }

                // try shrink y
                coord = 1;
                ny = spans[coord, 0] - 1; // new y
                if ((ny >= 0) && (distance > (point[coord] - (min[coord] + (ny + 1) * steps[coord]))))
                {
                    // it makes sense, check all cells with nx
                    for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                        for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{x, ny, z, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 0] = ny;
                }

                // try shrink z
                coord = 2;
                nz = spans[coord, 0] - 1; // new z
                if ((nz >= 0) && (distance > (point[coord] - (min[coord] + (nz + 1) * steps[coord]))))
                {
                    // it makes sense, check all cells with nx
                    for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
                        for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                            for (int t = spans[3, 0]; t <= spans[3, 1]; t++)
                            {
                                coords = new int[]{x, y, nz, t};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 0] = nz;
                }

                // try shrink t
                coord = 3;
                nt = spans[coord, 0] - 1; // new t
                if ((nt >= 0) && (distance > (point[coord] - (min[coord] + (nt + 1) * steps[coord]))))
                {
                    // it makes sense, check all cells with nx
                    for (int y = spans[1, 0]; y <= spans[1, 1]; y++)
                        for (int z = spans[2, 0]; z <= spans[2, 1]; z++)
                            for (int x = spans[0, 0]; x <= spans[0, 1]; x++)
                            {
                                coords = new int[]{x, y, z, nt};
                                if (cellMakesSense(coords, initCell, min, steps, point, distance))
                                {
                                    d = checkCell(coords, point, pi);
                                    if (d < distance) distance = d;
                                }
                            }
                    working = true;
                    spans[coord, 0] = nt;
                }
            }
            if (double.IsInfinity(distance))
            {
                throw new MveException(this, "Distance evaluated as infinity - probably there are no tetrahedra to test against.");
            }
            return (distance);
        }

		/// <summary>
		/// Performs the computation. May be very time-expensive. If no inputs are changed since the last run the it returns the last result immediately.
		/// </summary>
        public override void Execute()
        {
            if (paramsChanged || (!IsAllDataSame()) || (result == null))
            {
                int[] divs = {xDiv,yDiv,zDiv,tDiv};

                // get inputs
                UnstrGrid input = (UnstrGrid)GetInput("Evaluated mesh");
                UnstrGrid target = (UnstrGrid)GetInput("Target mesh");
                tetrahedra = target.GetCells(typeof(Tetrahedron))[0];
				points = target.Points;

                // create the grid
                PointND min;
                double[] steps;
                grid = CreateGrid(divs, target, tetrahedra, out min, out steps);

                // precompute data
                int edgeCount;
                precomputedData = PrecomputeData(tetrahedra, target.Points, out edgeCount);

                // create the new attribute
                bool pND = (input.Points.ElemType == typeof(PointND));
                if ((!pND) && (!(input.Points.ElemType == typeof(Point3D))))
                {
                    throw new MveException(this, "evaluated mesh must consist either of 4D PointND or Point3D.");
                }

                /*
                // determine obtuse tetrahedra
                obtuse = new bool[tetrahedra.Count];
                int obtuseCount = 0;
                for (int i = 0; i < tetrahedra.Count; i++)
                {
                    obtuse[i] = isTetrahedronObtuse((Tetrahedron)tetrahedra[i], target.Points);
                    if (obtuse[i]) obtuseCount++;
                }
                Debug.WriteLine("" + obtuseCount + " obstuse, " + (tetrahedra.Count - obtuseCount) + " non-obtuse");
                */
                 
                PointToTetrahedronDistanceEvaluator.initHash(target.Points.Count, edgeCount);
                UniformDataArray distances = new UniformDataArray(typeof(Scalar), input.Points.Count);
                checks = new int[tetrahedra.Count];
                for (int i = 0; i < distances.Count; i++)
                {
                    PointND p;
                    if (pND)
                    {
                        p = (PointND)input.Points[i];
                    }
                    else 
                    {
                        Point3D p3 = (Point3D)input.Points[i];
                        p = new PointND(p3.X, p3.Y, p3.Z, t);
                    }     
                    step = i+1;                    
                    double distance = distanceTo(p, divs, min, steps, i);
                    distances[i] = new Scalar(distance);
                    if ((i % 1000) == 0) 
                    {
                        ProgressInfo((float)i / distances.Count);
                        Debug.WriteLine(i);
                    }
                }

                // add the new attribute
                result = input.ShallowCopy();
                result.AddPointAttrs("distances", distances);

                // return result
                SetOutput("Resulting mesh", result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("Resulting mesh", result, true);
            }
        }
    }
}
