using System;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Text;
using System.Collections.Generic;
using System.ComponentModel;

namespace Examples
{
	/// <summary>
	/// Module for extraction of one frame from an animation represented by a 4D tetrahedral mesh. Only triangles within selected span of time are put to the output.
	/// </summary>
    [ModuleInfo("Libor Va", "Module for extraction of one frame from an animation represented by a 4D tetrahedral mesh. Only triangles within selected span of time are put to the output.", IconName="frameExtractor.ico")]
    [PortInfo("Mesh", "Input tetrahedral mesh, should be created by TetraMeshComposer or similar module.")]
    [PortInfo("TriangleMesh", "Output frame of the animation")]
    [PortInfo("TMin", "Start of the time span of the frame. The time span should be large enough to contain some triangles. If not connected to any input, then the value of the TMin property is used.")]
    [PortInfo("TMax", "End of the time span of the frame. The time span should be large enough to contain some triangles. If not connected to any input, then the value of the TMax property is used.")]
    class FrameExtractor: Module
    {
        double tmin = -0.02;
        double tmax = 0.02;

        bool paramsChanged = true;

        TriangleMesh result;

        /// <summary>
        /// Start of the time span. Has no effect if there's an input connected to the TMin port
        /// </summary>
        [Description("Start of the time span. Has no effect if there's an input connected to the TMin port")]
        public double TMin
        {
            get { return (tmin); }
            set
            {
                if (double.IsNaN(value))
                {
                    throw new MveException(this, "NaN passed as TMin.");
                }
                if (double.IsInfinity(value))
                {
                    throw new MveException(this, "Infinity passed as TMin.");
                }
                tmin = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// End of the time span. Has no effect if there's an input connected to the TMax port
        /// </summary>
        [Description("Start of the time span. Has no effect if there's an input connected to the TMax port")]
        public double TMax
        {
            get { return (tmax); }
            set
            {
                if (double.IsNaN(value))
                {
                    throw new MveException(this, "NaN passed as TMax.");
                }
                if (double.IsInfinity(value))
                {
                    throw new MveException(this, "Infinity passed as TMax.");
                }
                tmax = value;
                paramsChanged = true;
            }            
        }

        /// <summary>
        /// Initializes a new instance of the frame extractor, adds ports
        /// </summary>
        public FrameExtractor()
        {
            AddInPort("Mesh", typeof(UnstrGrid));
            AddInPort("TMin", typeof(Scalar), false);
            AddInPort("TMax", typeof(Scalar), false);
            AddOutPort("TriangleMesh", typeof(TriangleMesh));
        }

        private class TriangleEqualityComparer : EqualityComparer<Triangle>
        {
            int triangleCount = 100;

            public TriangleEqualityComparer(int count)
            {
                this.triangleCount = count;
            }

            public override bool Equals(Triangle x, Triangle y)
            {
                if (((x.V1 == y.V1) && (x.V2 == y.V2) && (x.V3 == y.V3))||
                ((x.V1 == y.V2) && (x.V2 == y.V3) && (x.V3 == y.V1))||
                ((x.V1 == y.V3) && (x.V2 == y.V1) && (x.V3 == y.V2)))
                    return (true);
                return false;
            }

            public override int GetHashCode(Triangle obj)
            {
                int index = 0;
                if (obj.V1 < obj.V2)
                    index += 4;
                if (obj.V2 < obj.V3)
                    index += 2;
                if (obj.V3 < obj.V1)
                    index++;
                switch (index)
                {
                    case 6: 
                        return (obj.V1 * triangleCount * triangleCount + obj.V2 * triangleCount + obj.V3);                        
                    case 4:
                        return (obj.V1 * triangleCount * triangleCount + obj.V3 * triangleCount + obj.V2);                        
                    case 2:
                        return (obj.V2 * triangleCount * triangleCount + obj.V1 * triangleCount + obj.V3);                        
                    case 3:
                        return (obj.V2 * triangleCount * triangleCount + obj.V3 * triangleCount + obj.V1);                        
                    case 1:
                        return (obj.V3 * triangleCount * triangleCount + obj.V2 * triangleCount + obj.V1);                        
                    case 5:
                        return (obj.V3 * triangleCount * triangleCount + obj.V1 * triangleCount + obj.V2);                        
                    default:
                        return (obj.V1 * triangleCount * triangleCount + obj.V3 * triangleCount + obj.V3);
                }
            }
        }

        /// <summary>
        /// Performs the extraction. If no parameters were changed since the last run, then the result of the last run is returned immediately with the DataSame flag set to true.
        /// </summary>
        public override void Execute()
        {
            if (paramsChanged ||
                (!IsDataSame("Mesh")) ||
                ((GetInPort("TMin").Link != null) && (!IsDataSame("TMin"))) ||
                ((GetInPort("TMax").Link != null) && (!IsDataSame("TMax"))) ||
                (result == null))
            {
                if (this.GetInPort("TMin").Link != null)
                {
                    Scalar s = (Scalar)GetInput("TMin");
                    tmin = s.Value;
                }

                if (this.GetInPort("TMax").Link != null)
                {
                    Scalar s = (Scalar)GetInput("TMax");
                    tmax = s.Value;
                }

                if (tmin > tmax)
                {
                    throw new MveException(this, "TMin specified is larger than TMax.");
                }

                List<int> indices = new List<int>();
                Dictionary<int, int> newIndices = new Dictionary<int, int>();

                UnstrGrid input = (UnstrGrid)GetInput("Mesh");

                if (input.Points.ElemType != typeof(PointND))
                {
                    throw new MveException(this, "Mesh expected to have 4D PointND vertices");
                }

                PointND p;
                for (int i = 0; i < input.Points.Count; i++)
                {
                    p = (PointND)input.Points[i];
                    if ((p[3] >= tmin) && (p[3] <= tmax))
                    {
                        indices.Add(i);
                        newIndices.Add(i, indices.Count - 1);
                    }
                }

                UniformDataArray newPoints = new UniformDataArray(typeof(Point3D), indices.Count);
                for (int i = 0; i < indices.Count; i++)
                {
                    p = (PointND)input.Points[indices[i]];
                    newPoints[i] = new Point3D(p[0], p[1], p[2]);
                }

                UniformDataArray tetrahedra = input.GetCells(typeof(Tetrahedron))[0];

                Dictionary<Triangle, object> triangles = new Dictionary<Triangle, object>(new TriangleEqualityComparer(3*indices.Count));

                bool c1, c2, c3, c4;
                int i1 = 0, i2 = 0, i3 = 0, i4 = 0;
                Triangle tri;
                int ti = 0;
                foreach (Tetrahedron t in tetrahedra)
                {
                    c1 = newIndices.TryGetValue(t.V1, out i1);
                    c2 = newIndices.TryGetValue(t.V2, out i2);
                    c3 = newIndices.TryGetValue(t.V3, out i3);
                    c4 = newIndices.TryGetValue(t.V4, out i4);

                    if (c1 && c2 && c3)
                    {
                        if (((PointND)input.Points[t.V4])[3] > ((PointND)input.Points[t.V1])[3])
                            tri = new Triangle(i1, i2, i3);
                        else
                            tri = new Triangle(i1, i3, i2);
                        if (!triangles.ContainsKey(tri))
                        {
                            triangles.Add(tri, null);
                        }
                    }

                    if (c3 && c2 && c4)
                    {
                        tri = new Triangle(i3, i2, i4);
                        if (!triangles.ContainsKey(tri))
                        {
                            triangles.Add(tri, null);
                        }
                    }

                    if (c1 && c3 && c4)
                    {
                        tri = new Triangle(i1, i3, i4);
                        if (!triangles.ContainsKey(tri))
                        {
                            triangles.Add(tri, null);
                        }
                    }

                    if (c4 && c2 && c1)
                    {
                        tri = new Triangle(i4, i2, i1);
                        if (!triangles.ContainsKey(tri))
                        {
                            triangles.Add(tri, null);
                        }
                    }
                    ti++;
                    if (ti % 1000 == 0)
                    {
                        ProgressInfo((float)ti / tetrahedra.Count);
                    }
                }
                
                UniformDataArray triangleArray = UniformDataArray.FromDictionaryKeys(triangles);

                result = new TriangleMesh();
                result.Points = newPoints;
                result.AddCells("triangles", triangleArray);

                foreach (string s in input.GetPointAttrsIds())
                {
                    UniformDataArray attrs = input.GetPointAttrs(s);
                    UniformDataArray newAttrs = new UniformDataArray(attrs.ElemType, indices.Count);
                    for (int i = 0; i < indices.Count; i++)
                        newAttrs[i] = attrs[indices[i]];
                    result.AddPointAttrs(s, newAttrs);
                }
                SetOutput("TriangleMesh",result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("TriangleMesh", result, true);
            }
        }
    }
}
