using System;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;

namespace Examples
{
	/// <summary>
	/// Summary description for NormalComputer.
	/// </summary>
    [ModuleInfo("Libor Va", "Module that computes vertex normals for a TriangleMesh. Normals are placed in a new point attribute of name specified in the AttrName property.", IconName = "NormalComputer.ico")]
	[PortInfo("TriangleMesh", "Input triangle mesh to be processed")]
	[PortInfo("Computed TriangleMesh", "Shallow copy of the input mesh with added attribute that represents vertex normals.")]
	public class NormalComputer:Module
	{
		TriangleMesh output = null;

		bool paramsChanged = true;

		string attrName = "normals";

        /// <summary>
        /// Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.
        /// </summary>
        [Description("Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.")]
		public string AttrName 
		{
			get{return(attrName);}
			set
			{
				if (value!="")
				{
					attrName = value;
					paramsChanged = true;
				}
				else
				{
					throw new MveException(this, "Attribute name for normals cannot be an empty strng.");
				}
			}
		}

        /// <summary>
        /// Default constructor.
        /// </summary>
		public NormalComputer()
		{
			AddInPort("TriangleMesh",typeof(TriangleMesh));
			AddOutPort("Computed TriangleMesh",typeof(TriangleMesh));
		}

        /// <summary>
        /// Computes the normals of the mesh. If the mesh is unchanged, and no properties were changed since the last run, then the result of the last run is returned with dataSame set to true.
        /// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(!IsAllDataSame())||(output == null))
			{
				TriangleMesh input = (TriangleMesh)GetInput("TriangleMesh");
				UniformDataArray triangles = (input.GetCells(typeof(Triangle)))[0];

				UniformDataArray normals = new UniformDataArray(typeof(Normal3D),input.Points.Count);
				for (int i = 0;i<normals.Count;i++)
					normals[i] = new Normal3D(0,0,0);

				int pos = 0;
				foreach(Triangle t in triangles)
				{
					Point3D v1 = (Point3D)input.Points[t.V2]-(Point3D)input.Points[t.V1];
					Point3D v2 = (Point3D)input.Points[t.V3]-(Point3D)input.Points[t.V1];
					Normal3D normal = (Normal3D)v1.CrossProduct(v2);
                    normal.Normalize();
					normals[t.V1] = normal + (Normal3D)normals[t.V1];
					normals[t.V2] = normal + (Normal3D)normals[t.V2];
					normals[t.V3] = normal + (Normal3D)normals[t.V3];
					pos++;
					if ((pos%1000) == 0)
						ProgressInfo((float)pos/triangles.Count);
				}

				for (int i = 0;i<normals.Count;i++)
				{
					Normal3D normal = (Normal3D)normals[i];
					normal.Normalize();
					normals[i] = normal;
				}
                output = (TriangleMesh)input.ShallowCopy();
				output.AddPointAttrs(attrName,normals);
				SetOutput("Computed TriangleMesh", output);
				paramsChanged = false;
			}
			else
			{
				SetOutput("Computed TriangleMesh", output, true);
			}
		}
	}
}
