using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Examples
{
	/// <summary>
	/// A class used for precomputed data about a tetrahedron
	/// </summary>
    class TetrahedronPrecomputedData
    {
        public PointND v1, v2, v3, n;
        public double tetrahedronDet, nDotN;
        public int e01, e02, e03, e12, e13, e23;
    }    

	/// <summary>
	/// A specialized module used to evaluate distance from a point to a 4D tetrahedron
	/// </summary>
	[ModuleInfo("Libor Va", "A specialized module used to evaluate distance from a point to a 4D tetrahedron.")]
	[PortInfo("Point", "A PointND to be evaluated")]
	[PortInfo("Tetrahedron", "A 4-element MveArray of PointND elements that represent the tetrahedron")]
	[PortInfo("Distance", "A MveScalar representing the evaluated distance value")]
    class PointToTetrahedronDistanceEvaluator:Module
    {        
        private static int[] pointHashKeys;        
        private static int[] edgeHashKeys;
        private static bool[] edgeHashValues;

		private double result;
		private bool resultSet = false;

		/// <summary>
		/// helper function that initializes internal data structures.
		/// </summary>
		/// <param name="pointCount">Number of points that will be evaluated</param>
		/// <param name="edgeCount">Number of edges of the target mesh</param>
        public static void initHash(int pointCount, int edgeCount)
        {
            pointHashKeys = new int[pointCount];
            edgeHashKeys = new int[edgeCount];
            edgeHashValues = new bool[edgeCount];
            for (int i = 0;i<pointCount;i++)
                pointHashKeys[i] = -1;

            for (int i = 0; i < edgeCount; i++)
                edgeHashKeys[i] = -1;
            
        }        

		/// <summary>
		/// Default constructor. Adds input and output ports.
		/// </summary>
        public PointToTetrahedronDistanceEvaluator()
        {
            AddInPort("Point", typeof(PointND));
            AddInPort("Tetrahedron", typeof(UniformDataArray));
            AddOutPort("Distance", typeof(Scalar));
        }        

        private static double[] gauss(double[,] matrix)
        {
            int l = matrix.GetLength(0);

            // forward run
            for (int c = 0;c<(l-1);c++)
                for (int r = (c+1);r<l;r++)
                {
                    if (matrix[r, c] != 0)
                    {
                        double m = matrix[c, c] / matrix[r, c];
                        for (int i = (c + 1); i < (l + 1); i++)
                        {
                            matrix[r, i] *= m;
                            matrix[r, i] -= matrix[c, i];
                        }
                        matrix[r, c] = 0;
                    }
                }
            // backward run
            for (int c = (l - 1); c > 0; c--)
            {
                for (int r = (c - 1); r >= 0; r--)
                {
                    if (matrix[r, c] != 0)
                    {
                        double m = matrix[c, c] / matrix[r, c];
                        for (int i = (c - 1); i < (l + 1); i++)
                        {
                            matrix[r, i] *= m;
                            matrix[r, i] -= matrix[c, i];
                        }
                    }
                }
            }

            double[] result = new double[l];
            for (int i = 0; i < l; i++)
                result[i] = matrix[i, l] / matrix[i, i];
            return (result);
        }

		/// <summary>
		/// helper function for determining a 4x4 matrix determinant
		/// </summary>
		/// <param name="c1">1. column of the matrix</param>
		/// <param name="c2">2. column of the matrix</param>
		/// <param name="c3">3. column of the matrix</param>
		/// <param name="c4">4. column of the matrix</param>
		/// <returns>determinant value</returns>
        public static double det4(double[] c1, double[] c2, double[] c3, double[] c4)
        {
            double sd1 = c2[1] * c3[2] * c4[3] + c2[2] * c3[3] * c4[1] + c2[3] * c3[1] * c4[2] - c2[3] * c3[2] * c4[1] - c2[2] * c3[1] * c4[3] - c2[1] * c3[3] * c4[2];
            double sd2 = c1[1] * c3[2] * c4[3] + c1[2] * c3[3] * c4[1] + c1[3] * c3[1] * c4[2] - c1[3] * c3[2] * c4[1] - c1[2] * c3[1] * c4[3] - c1[1] * c3[3] * c4[2];
            double sd3 = c1[1] * c2[2] * c4[3] + c1[2] * c2[3] * c4[1] + c1[3] * c2[1] * c4[2] - c1[3] * c2[2] * c4[1] - c1[2] * c2[1] * c4[3] - c1[1] * c2[3] * c4[2];
            double sd4 = c1[1] * c2[2] * c3[3] + c1[2] * c2[3] * c3[1] + c1[3] * c2[1] * c3[2] - c1[3] * c2[2] * c3[1] - c1[2] * c2[1] * c3[3] - c1[1] * c2[3] * c3[2];
            double result = c1[0] * sd1 - c2[0] * sd2 + c3[0] * sd3 - c4[0]*sd4;
            return (result);
        }

        private static double det3(double[] c1, double[] c2, double[] c3)
        {
            double result = c1[0] * c2[1] * c3[2] + c1[1] * c2[2] * c3[0] + c1[2]*c2[0]*c3[1] - c1[2] * c2[1] * c3[0] - c1[1] * c2[0] * c3[2] - c1[0] * c2[2] * c3[1];
            return (result);
        }

        private static double[] sarus3(double[] c1, double[] c2, double[] c3, double[] r)
        {
            double denominator = det3(c1, c2, c3);
            if (Math.Abs(denominator) < 0.0000001) return (null);
            double nom1 = det3(r, c2, c3);
            double nom2 = det3(c1, r, c3);
            double nom3 = det3(c1, c2, r);
            double[] result = new double[3];
            result[0] = nom1 / denominator;
            result[1] = nom2 / denominator;
            result[2] = nom3 / denominator;
            return (result);
        }

        private static double[] sarus4(double[] c1, double[] c2, double[] c3, double[] c4, double[] r)
        {
            double denominator = det4(c1, c2, c3, c4);
            if (Math.Abs(denominator) < Scalar.Epsilon) return (null);
            double nom1 = det4(r, c2, c3, c4);
            double nom2 = det4(c1, r, c3, c4);
            double nom3 = det4(c1, c2, r, c4);
            double nom4 = det4(c1, c2, c3, r);
            double[] result = new double[4];
            result[0] = nom1 / denominator;
            result[1] = nom2 / denominator;
            result[2] = nom3 / denominator;
            result[3] = nom4 / denominator;
            return (result);
        }

        private static double[] sarus4(double[] c1, double[] c2, double[] c3, double[] c4, double[] r, TetrahedronPrecomputedData tpd)
        {
            double denominator = tpd.tetrahedronDet;
            if (Math.Abs(denominator) < Scalar.Epsilon) return (null);
            double nom1 = det4(r, c2, c3, c4);
            double nom2 = det4(c1, r, c3, c4);
            double nom3 = det4(c1, c2, r, c4);
            double nom4 = det4(c1, c2, c3, r);
            double[] result = new double[4];
            result[0] = nom1 / denominator;
            result[1] = nom2 / denominator;
            result[2] = nom3 / denominator;
            result[3] = nom4 / denominator;
            return (result);
        }

		/// <summary>
		/// Helper function determining wether a point is projected on a tetrahedron.
		/// </summary>
		/// <param name="point">point to be projected</param>
		/// <param name="distance">distance to the tetrahedron }only if projection is on the tetrahedron</param>
		/// <param name="tetrahedron">tetrahedronto onto which to project</param>
		/// <returns>true if projection lies on the tetrahedron, false otherwise</returns>
        public static bool isProjectionOnTetrahedron(PointND point, out double distance, params PointND[] tetrahedron)
        {
            PointND v1 = tetrahedron[1] - tetrahedron[0];
            PointND v2 = tetrahedron[2] - tetrahedron[0];
            PointND v3 = tetrahedron[3] - tetrahedron[0];            

            // project to tetrahedron hyperplane (space)
            PointND n = PointND.crossProduct(v1, v2, v3);

            PointND r = point - tetrahedron[0];
            double[] coefs = sarus4(v1.Values, v2.Values, v3.Values, n.Values, r.Values);

            if ((coefs[0] >= 0) && (coefs[1] >= 0) && (coefs[2] >= 0) && ((coefs[0] + coefs[1] + coefs[2]) <= 1))
            {
                if (Math.Abs(coefs[3]) < Scalar.Epsilon)
                {
                    // point is within tetrahedron
                    distance = 0;
                    return (true);
                }
                // projection is inside                    
                distance = Math.Abs(n.dotProduct(n) * coefs[3]*coefs[3]);
                return (true);
            }
            distance = double.PositiveInfinity;            
            return (false);
        }

		/// <summary>
		/// Helper function determining wether a point is projected on a tetrahedron.
		/// </summary>
		/// <param name="point">Point to be projected</param>
		/// <param name="distance">distance to the tetrahedron (only if projection lies on it)</param>
		/// <param name="tpd">precomputed data about the tetrahedron to use.</param>
		/// <param name="tetrahedron">tetrahedron to project to.</param>
		/// <returns>true if projection lies on the tetrahedron, false otherwise</returns>
        private static bool isProjectionOnTetrahedron(PointND point, out double distance, TetrahedronPrecomputedData tpd, params PointND[] tetrahedron)
        {
            /*PointND v1 = tetrahedron[1] - tetrahedron[0];
            PointND v2 = tetrahedron[2] - tetrahedron[0];
            PointND v3 = tetrahedron[3] - tetrahedron[0];

            // project to tetrahedron hyperplane (space)
            PointND n = crossProduct(v1, v2, v3);*/

            PointND r = point - tetrahedron[0];
            double[] coefs = sarus4(tpd.v1.Values, tpd.v2.Values, tpd.v3.Values, tpd.n.Values, r.Values, tpd);

            if ((coefs[0] >= 0) && (coefs[1] >= 0) && (coefs[2] >= 0) && ((coefs[0] + coefs[1] + coefs[2]) <= 1))
            {
                if (Math.Abs(coefs[3]) < Scalar.Epsilon)
                {
                    // point is within tetrahedron
                    distance = 0;
                    return (true);
                }
                // projection is inside                    
                distance = Math.Abs(tpd.nDotN * coefs[3]*coefs[3]);
                return (true);
            }
            distance = double.PositiveInfinity;
            return (false);
        }        

		/// <summary>
		/// Helper function that determines wether a point projection lies on a tetrahedron face.
		/// </summary>
		/// <param name="point">Point to be projected</param>
		/// <param name="distanceSq">Square of the projection distance</param>
		/// <param name="triangle">Vertices of the face</param>
		/// <returns>true if projection lies on the face, false otherwise</returns>
        public static bool isProjectionOnFace(PointND point, out double distanceSq, params PointND[] triangle)
        {
            PointND v1 = (triangle[1] - triangle[0]);
            PointND v2 = (triangle[2] - triangle[0]);
            PointND v3 = (point - triangle[0]);

            PointND binormal = PointND.crossProduct(v1, v2, v3);

            PointND n = PointND.crossProduct(v1, v2, binormal);                        
            
            double[] coefs = sarus4(v1.Values, v2.Values, n.Values, binormal.Values, v3.Values);

            if (coefs == null)
            {
                distanceSq = double.PositiveInfinity;
                return (false);
            }

            if ((coefs[0] >= 0) && (coefs[1] >= 0) && ((coefs[0] + coefs[1]) <= 1))
            {
                // projection is on the plane, return distance to the plane                    
				distanceSq = Math.Abs(n.dotProduct(n) * coefs[2] * coefs[2]);                
                return(true);
            }
			distanceSq = double.PositiveInfinity;            
            return (false);
        }

		/// <summary>
		/// A helper function that determines wether a point is projected onto a line segment
		/// </summary>
		/// <param name="point">Point to be projected</param>
		/// <param name="distanceSq">Square of the distance to the line segment }only if the projection lies on the line segment.</param>
		/// <param name="edge">Segment vertices</param>
		/// <returns>true if projection lies on the line segment, false otherwise</returns>
        public static bool isProjectionOnEdge(PointND point, out double distanceSq, params PointND[] edge)
        {
            PointND v1 = edge[1] - edge[0];

            PointND v2 = point - edge[0];

            double lengthSq = v1.dotProduct(v1);

            double dp = v1.dotProduct(v2);

            double t = dp / lengthSq;

            if ((t >= 0) && (t <= 1))
            {
                // projection lies on the edge                    
                double c = v2.dotProduct(v2);
                distanceSq = c - (lengthSq * t * t);
                return (true);
            }
            distanceSq = double.PositiveInfinity;
            return (false);
        }

		/// <summary>
		/// Determines the distance to a tetrahedron
		/// </summary>
		/// <param name="point">Point to evaluate</param>
		/// <param name="tetrahedron">tetrahedron vertices</param>
		/// <returns>Distance</returns>
        public static double EvaluateDistance(PointND point, params PointND[] tetrahedron)
        {
            double dist = double.PositiveInfinity;
            double nd;
            if (isProjectionOnTetrahedron(point, out dist,tetrahedron))
                return(dist);

            int f0 = 0, f1 = 0, f2 = 0, f3 = 0;

            // edges

            if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[1]))
            {
                if (nd < dist) dist = nd;
                f2++;
                f3++;                
            }
            if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[2]))
            {
                if (nd < dist) dist = nd;
                f1++;
                f3++;
            }
            if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[3]))
            {
                if (nd < dist) dist = nd;
                f1++;
                f2++;
            }
            if (isProjectionOnEdge(point, out nd, tetrahedron[1], tetrahedron[2]))
            {
                if (nd < dist) dist = nd;
                f0++;
                f3++;
            }
            if (isProjectionOnEdge(point, out nd, tetrahedron[1], tetrahedron[3]))
            {
                if (nd < dist) dist = nd;
                f0++;
                f2++;
            }
            if (isProjectionOnEdge(point, out nd, tetrahedron[2], tetrahedron[3]))
            {
                if (nd < dist) dist = nd;
                f0++;
                f1++;
            }

            // vertices

            PointND v = point - tetrahedron[0];
            nd = v.dotProduct(v);
            if (nd < dist) dist = nd;

            v = point - tetrahedron[1];
            nd = v.dotProduct(v);
            if (nd < dist) dist = nd;

            v = point - tetrahedron[2];
            nd = v.dotProduct(v);
            if (nd < dist) dist = nd;

            v = point - tetrahedron[3];
            nd = v.dotProduct(v);
            if (nd < dist) dist = nd;                                   

            // faces
            if ((f3>1)&&(isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[1], tetrahedron[2])))
            {
                if (nd<dist) dist = nd;
            }
            if ((f2>1)&&(isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[1], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }
            if ((f1>1)&&(isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[2], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }
            if ((f0>1)&&(isProjectionOnFace(point, out nd, tetrahedron[1], tetrahedron[2], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }

			dist = Math.Sqrt(dist);

            return (dist);
        }

		/// <summary>
		/// Helper function that determines the distance to a tetrahedron
		/// </summary>
		/// <param name="point">Point to evaluate</param>
		/// <param name="tpd">precomputed data about the tetrahedron</param>
		/// <param name="pi">index of the point</param>
		/// <param name="t0">index of the 1. vertex of the tetrahedron</param>
		/// <param name="t1">index of the 2. vertex of the tetrahedron</param>
		/// <param name="t2">index of the 3. vertex of the tetrahedron</param>
		/// <param name="t3">index of the 4. vertex of the tetrahedron</param>
		/// <param name="tetrahedron">tetrahedron vertices</param>
		/// <returns>Distance to the tetrahedron</returns>
        public static double EvaluateDistance(PointND point, TetrahedronPrecomputedData tpd, int pi, int t0, int t1, int t2, int t3, params PointND[] tetrahedron)
        {
            double dist = double.PositiveInfinity;
            double nd;
            if (isProjectionOnTetrahedron(point, out dist, tpd, tetrahedron))
                return (Math.Sqrt(dist));

            int f0 = 0, f1 = 0, f2 = 0, f3 = 0;

            // edges            
            
            if (edgeHashKeys[tpd.e01] == pi)
            {
                if (edgeHashValues[tpd.e01])
                {
                    f2++;
                    f3++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e01] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[1]))
                {
                    edgeHashValues[tpd.e01] = true;                    
                    f2++;
                    f3++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e01] = false;
                }
            }

            if (edgeHashKeys[tpd.e02] == pi)
            {
                if (edgeHashValues[tpd.e02])
                {
                    f1++;
                    f3++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e02] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[2]))
                {
                    edgeHashValues[tpd.e02] = true;
                    f1++;
                    f3++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e02] = false;
                }
            }


            if (edgeHashKeys[tpd.e03] == pi)
            {
                if (edgeHashValues[tpd.e03])
                {
                    f1++;
                    f2++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e03] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[0], tetrahedron[3]))
                {
                    edgeHashValues[tpd.e03] = true;
                    f1++;
                    f2++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e03] = false;
                }
            }

            if (edgeHashKeys[tpd.e12] == pi)
            {
                if (edgeHashValues[tpd.e12])
                {
                    f0++;
                    f3++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e12] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[1], tetrahedron[2]))
                {
                    edgeHashValues[tpd.e12] = true;
                    f0++;
                    f3++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e12] = false;
                }
            }

            if (edgeHashKeys[tpd.e13] == pi)
            {
                if (edgeHashValues[tpd.e13])
                {
                    f0++;
                    f2++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e13] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[1], tetrahedron[3]))
                {
                    edgeHashValues[tpd.e13] = true;
                    f0++;
                    f2++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e13] = false;
                }
            }

            if (edgeHashKeys[tpd.e23] == pi)
            {
                if (edgeHashValues[tpd.e23])
                {
                    f0++;
                    f1++;
                }
            }
            else
            {
                edgeHashKeys[tpd.e23] = pi;
                if (isProjectionOnEdge(point, out nd, tetrahedron[2], tetrahedron[3]))
                {
                    edgeHashValues[tpd.e23] = true;
                    f0++;
                    f1++;
                    if (nd < dist) dist = nd;
                }
                else
                {
                    edgeHashValues[tpd.e23] = false;
                }
            }

            // vertices
            if (pointHashKeys[t0] != pi)
            {
                PointND v = point - tetrahedron[0];
                nd = v.dotProduct(v);
                pointHashKeys[t0] = pi;
                if (nd < dist) dist = nd;                
            }                        

            if (pointHashKeys[t1] != pi)
            {
                PointND v = point - tetrahedron[1];
                nd = v.dotProduct(v);
                pointHashKeys[t1] = pi;
                if (nd < dist) dist = nd;
            }                        

            if (pointHashKeys[t2] != pi)
            {
                PointND v = point - tetrahedron[2];
                nd = v.dotProduct(v);
                pointHashKeys[t2] = pi;
                if (nd < dist) dist = nd;
            }                        

            if (pointHashKeys[t3] != pi)
            {
                PointND v = point - tetrahedron[3];
                nd = v.dotProduct(v);
                pointHashKeys[t3] = pi;
                if (nd < dist) dist = nd;
            }                                   

            // faces
            if ((f3 > 1) && (isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[1], tetrahedron[2])))
            {
                if (nd < dist) dist = nd;
            }
            if ((f2 > 1) && (isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[1], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }
            if ((f1 > 1) && (isProjectionOnFace(point, out nd, tetrahedron[0], tetrahedron[2], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }
            if ((f0 > 1) && (isProjectionOnFace(point, out nd, tetrahedron[1], tetrahedron[2], tetrahedron[3])))
            {
                if (nd < dist) dist = nd;
            }

			dist = Math.Sqrt(dist);

            return (dist);
        }

        private static double[] leastSquares2(PointND p1, PointND p2, PointND p3)
        {
            double aa = p1.dotProduct(p1);
            double ab = p1.dotProduct(p2);
            double bb = p2.dotProduct(p2);
            double ac = p1.dotProduct(p3);
            double bc = p2.dotProduct(p3);
            double denom = aa * bb - ab * ab;
            double nom1 = ac * bb - ab * bc;
            double nom2 = aa * bc - ac * ab;
            double[] result = new double[] { nom1 / denom, nom2 / denom, 0};
            return (result);
        }

		/// <summary>
		/// Distance to a non-obtuse tetrahedron. Depricated
		/// </summary>
		/// <param name="point">point to evaluate</param>
		/// <param name="tetrahedron">tetrahedron vertices</param>
		/// <returns>distance</returns>
        public static double EvaluateDistanceNO(PointND point, params PointND[] tetrahedron)
        {
            double result;

            PointND v1 = tetrahedron[1] - tetrahedron[0];
            PointND v2 = tetrahedron[2] - tetrahedron[0];
            PointND v3 = tetrahedron[3] - tetrahedron[0];            

            // project to tetrahedron hyperplane (space)
            PointND n = PointND.crossProduct(v1, v2, v3);

            PointND r = point - tetrahedron[0];
            double[] coefs = sarus4(v1.Values, v2.Values, v3.Values, n.Values, r.Values);            

            if ((coefs[0] >= 0) && (coefs[1] >= 0) && (coefs[2] >= 0) && ((coefs[0] + coefs[1] + coefs[2]) <= 1))
            {
                if (Math.Abs(coefs[3])<Scalar.Epsilon)
                {
                    // point is within tetrahedron                    
                    return(0);
                }
                // projection is inside                    
                result = Math.Abs(n.Size() * coefs[3]);                
                return(result);
            }

            PointND p1, p2, p3;            

            if (coefs[0] < 0)
            {
                p1 = tetrahedron[0];
                p2 = tetrahedron[2];
                p3 = tetrahedron[3];
            }
            else if (coefs[1] < 0)
            {
                p1 = tetrahedron[0];
                p2 = tetrahedron[1];
                p3 = tetrahedron[3];
            }
            else if (coefs[2] < 0)
            {
                p1 = tetrahedron[0];
                p2 = tetrahedron[1];
                p3 = tetrahedron[2];
            }
            else
            {
                // coefs sum is larger than 1
                p1 = tetrahedron[1];
                p2 = tetrahedron[2];
                p3 = tetrahedron[3];
            }

            v1 = (p2 - p1);
            v2 = (p3 - p1);
            v3 = (point - p1);

            PointND binormal = PointND.crossProduct(v1, v2, v3);

            if (binormal.dotProduct(binormal) < Scalar.Epsilon)
            {
                // point lies in the plane of a face, let's see wether it's inside
                coefs = leastSquares2(v1, v2, v3);
            }
            else
            {

                n = PointND.crossProduct(v1, v2, binormal);                

                coefs = sarus3(v1.Values, v2.Values, n.Values, v3.Values);

                if (coefs == null)
                {
                    coefs = sarus4(v1.Values, v2.Values, n.Values, binormal.Values, r.Values);
                    
                }
            }

            if ((coefs[0] >= 0) && (coefs[1] >= 0) && ((coefs[0] + coefs[1]) <= 1))
            {
                // projection is on the plane, return distance to the plane                    
                result = Math.Abs(n.Size() * coefs[2]);
                return (result);
            }
        

            PointND l1, l2;

            if (coefs[0] < 0)
            {
                l1 = p1;
                l2 = p3;
            }
            else if (coefs[1] < 0)
            {
                l1 = p1;
                l2 = p2;
            }
            else
            {
                //Sum greater than 1
                l1 = p2;
                l2 = p3;
            }

            v1 = l2 - l1;

            v2 = point - l1;

            double lengthSq = v1.dotProduct(v1);

            //v1.Normalize();

            double dp = v1.dotProduct(v2);

            double t = dp / lengthSq;

            if ((t >= 0) && (t <= 1))
            {
                // projection lies on the edge                    
                double c = v2.dotProduct(v2);
                result = Math.Sqrt(c - (lengthSq * t * t));                
                return(result);
            }
            else if (t < 0)
            {
                PointND p = point - l1;
                result = p.Size();                
                return(result);
            }
            else
            {
                PointND p = point - l2;
                result = p.Size();                
                return(result);
            }            
        }

		/// <summary>
		/// performs the evaluation
		/// </summary>
        public override void Execute()
        {
			if ((!IsAllDataSame())||(!resultSet))
			{
				// get data

				UniformDataArray tetrahedron = (UniformDataArray)GetInput("Tetrahedron");
				if (tetrahedron.ElemType != typeof(PointND))
				{
					throw new MveException(this, "Distance evlauator expects an UniformDataArray instance containing PointNDs.");
				}
				if (tetrahedron.Count != 4)
				{
					throw new MveException(this, "Distance evlauator expects an UniformDataArray instance containing exactly four PointNDs.");
				}

				PointND[] tetraPoints = new PointND[4];
				for (int i = 0; i < 4; i++)
				{
					tetraPoints[i] = (PointND)tetrahedron[i];
					if (tetraPoints[i].Length != 4)
					{
						throw new MveException(this, "Distance evaluator expects a tetrahedron in 4D, i.e. defined by four-dimensional PointND instances.");
					}
				}

				PointND point = (PointND)GetInput("Point");

				if (point.Length != 4)
				{
					throw new MveException(this, "Distance evaluator expects a 4D point, i.e. a four-dimensional instance of PointND.");
				}

				result = EvaluateDistance(point, tetraPoints);
				SetOutput("Distance", new Scalar(result));
				resultSet = true;
			}
			else
			{
				SetOutput("Distance", new Scalar(result));
			}
        }
    }
}
