using System;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Examples
{
    /// <summary>
    /// A module that composes a series of subsequent constant topology meshes into a tetrahedral mesh in hyperspace.
    /// </summary>
	[ModuleInfo("Libor Va","A module that composes a series of subsequent constant topology meshes into a tetrahedral mesh in hyperspace.", IconName = "tetraMeshComposer.ico")]
	[PortInfo("Mesh", "A Mesh structure that represents one of input meshes for time frame specified by the Frame number port.")]
	[PortInfo("Frame number","MveInt that specifies what frame is needed at the Mesh input port.")]
	[PortInfo("TetraMesh","A resulting tetrahedral mesh.")]
	public class TetraMeshComposer:Module
	{
        /// <summary>
        /// Default constructor
        /// </summary>
		public TetraMeshComposer()
		{
			AddInPort("Mesh",typeof(Mesh),true,false);
			AddOutPort("TetraMesh",typeof(UnstrGrid));
			AddOutPort("Frame number",typeof(MveInt),false);
		}

		bool paramsChanged = true;
		bool recompute = false;
		int meshCount = 2;
		UnstrGrid output = null;
		double frameDistance = 1/25.0;

        /// <summary>
        /// Gets or sets the distance between frames in the additional dimension. Set to time span between frames.
        /// </summary>
        [Description("Gets or sets the distance between frames in the additional dimension. Set to time span between frames.")]
		public double FrameDistance
		{
			get{return(frameDistance);}
			set
			{
				if ((double.IsInfinity(value))||(double.IsNaN(value)))
				{
					throw new MveException(this, "Frame distance must be a number.");
				}
				else
				{
					frameDistance = value;
				}
			}
		}

        /// <summary>
        /// Determines wether the result should or not be recomputed when result from previous run is available.
        /// </summary>
        [Description("Determines wether the result should or not be recomputed when result from previous run is available.")]
		public bool Recompute
		{
			get {return(recompute);}
			set {recompute = value;}
		}

        /// <summary>
        /// Gets or sets the number of meshes that will be composed into a time-space object. Must be greater or equal to two.
        /// </summary>
        [Description("Gets or sets the number of meshes that will be composed into a time-space object. Must be greater or equal to two.")]
		public int MeshCount
		{
			get{return(meshCount);}
			set
			{
				if (value>1)
				{
					meshCount = value;
					paramsChanged = true;
				}
				else
				{
					throw new MveException(this, "Tetrahedral mesh cannot be composed of less than 2 frames.");
				}
			}
		}

        /// <summary>
        /// Performs the time-space object construction. Sets the Frame number output to the required frame index and asks for data at the Mesh input for each frame.
        /// Number of demanded frames is determined by the MeshCount property.
        /// Finally, puts the result to the output.
        /// When result from previous run is available and Recompute property is set to false, then it is put to output immediately and dataSame is set to true.
        /// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(recompute)||(output == null))
			{
				output = new UnstrGrid();				

				SetOutput("Frame number",new MveInt(0),false);

				UpdateInput("Mesh");
				Mesh input = (Mesh)GetInput("Mesh");

				int pointCount = input.Points.Count;

				UniformDataArray points = new UniformDataArray(typeof(PointND),input.Points.Count*meshCount);

				UniformDataArray triangles = input.GetCells(typeof(Triangle))[0];

				int triangleCount = triangles.Count;

				UniformDataArray tetrahedra = new UniformDataArray(typeof(Tetrahedron),3*triangles.Count*(meshCount-1));

				for (int i = 0;i<input.Points.Count;i++)
				{
					Point3D point = (Point3D)input.Points[i];
					points[i] = new PointND(new double[]{point.X,point.Y,point.Z,0});
				}

				// compute one level of tetrahedra
				int index = 0;
				foreach(Triangle t in triangles)
				{
					if (t.V1>t.V2)
					{
						if (t.V1>t.V3) // V1 is largest
						{
							tetrahedra[index] = new Tetrahedron(t.V1, t.V2, t.V3, t.V1+pointCount);
							if (t.V2<t.V3) // V2 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V2);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V3, t.V1+pointCount, t.V3 + pointCount);
							}
							else // V3 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V3);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V3, t.V1+pointCount, t.V2 + pointCount);
 							}
						}
						else // V3 is largest
						{
							tetrahedra[index] = new Tetrahedron(t.V1, t.V2, t.V3, t.V3+pointCount);
							if (t.V1<t.V2) // V1 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V1);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V1, t.V2+pointCount, t.V3 + pointCount);
							}
							else // V2 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V2);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V1, t.V1+pointCount, t.V3 + pointCount);
							}
						}
					}
					else //V1<V2
					{
						if (t.V2>t.V3) // V2 is largest
						{
							tetrahedra[index] = new Tetrahedron(t.V1, t.V2, t.V3, t.V2+pointCount);
							if (t.V1<t.V3) // V1 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V1);
								tetrahedra[index+2] = new Tetrahedron(t.V3, t.V1, t.V2+pointCount, t.V3 + pointCount);
							}
							else // V3 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V3);
								tetrahedra[index+2] = new Tetrahedron(t.V1, t.V3, t.V1+pointCount, t.V2 + pointCount);
							}
						}
						else // V3 is largest
						{
							tetrahedra[index] = new Tetrahedron(t.V1, t.V2, t.V3, t.V3+pointCount);
							if (t.V1<t.V2) // V1 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V1);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V1, t.V2+pointCount, t.V3 + pointCount);
							}
							else // V2 is smallest
							{
								tetrahedra[index+1] = new Tetrahedron(t.V1+pointCount, t.V3+pointCount, t.V2+pointCount, t.V2);
								tetrahedra[index+2] = new Tetrahedron(t.V2, t.V1, t.V1+pointCount, t.V3 + pointCount);
							}
						}
					}
					index+=3;
				}
				
				for (int i = 1;i<meshCount;i++)
				{
					SetOutput("Frame number",new MveInt(i),false);
					UpdateInput("Mesh");
					input = (Mesh)GetInput("Mesh");
					for (int j = 0;j<input.Points.Count;j++)
					{
						Point3D point = (Point3D)input.Points[j];
						points[i*pointCount+j] = new PointND(new double[]{point.X,point.Y,point.Z,i*frameDistance});
					}

					// copy tetrahedal topology
					if (i<(meshCount-1))
					{
						for (int j = 0;j<(triangleCount*3);j++)
						{
							Tetrahedron t = (Tetrahedron)tetrahedra[j];
							t.V1+=i*pointCount;
							t.V2+=i*pointCount;
							t.V3+=i*pointCount;
							t.V4+=i*pointCount;
							tetrahedra[j+(triangleCount*3)*i] = t;
						}
					}
				}

				output.Points = points;
				output.AddCells("Tetrahedra",tetrahedra);
				SetOutput("TetraMesh",output,false);
				paramsChanged = false;
			}
			else
			{
				SetOutput("TetraMesh",output,true);
			}
		}
	}
}
