using System;
using System.Collections.Generic;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Examples
{
    /// <summary>
    /// A Module that produces 3D triangle meshes by slicing a 4D tetrahedral mesh by hyperplane (i.e. 3D space) defined by input vector.
    /// </summary>
	[ModuleInfo("Libor Va","This module produces 3D triangle meshes by slicing a 4D tetrahedral mesh by hyperplane (i.e. 3D space) defined by input vector.")]
	[PortInfo("TetraMesh", "A tetrahedral mesh to be sliced")]
	[PortInfo("Plane", "A vectorND instance that represents the slicing plane. The plane should be represented by 5 values that form the implicit equation of a 4D hyperplane.")]
	[PortInfo("TriangleMesh", "Resulting triangle mesh.")]
	public class TetraMeshSlicer:Module
	{
        private class LineEqualityComparer : IEqualityComparer<Line>
        {
            public static int pointCount = 1;

            public bool Equals(Line l1, Line l2)
            {                
                if ((l1.V1 == l2.V1) && (l1.V2 == l2.V2)) return (true);
                else return (false);
            }

            public int GetHashCode(Line l)
            {                
                return (l.V1 * pointCount + l.V2);
            }
        }


		private TriangleMesh output;

		private bool paramsChanged = true;

		private UniformDataArray sourcePoints;
		private UniformDataArray tetrahedra;
		private Dictionary<Line,int> edges;
		private List<Triangle> triangles;
		private List<Point3D> points;

		private int maxIndex = 0; // index of the most similar axis

		private Transform4D projMatrix;
		private PointND plane;

		private bool[] evals;

        private string triangleCellsName = "Triangles";

        /// <summary>
        /// Gets or sets the identifier of the output cells that will contain triangles.
        /// </summary>
        [Description("Gets or sets the identifier of the output cells that will contain triangles.")]
        public string TriangleCellsName
        {
            get { return triangleCellsName; }
            set 
            {
                if (value != "")
                {
                    triangleCellsName = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Triangle cells identifier cannot be an empty string.");
                }
            }
        }
	

        /// <summary>
        /// Default constructor.
        /// </summary>
		public TetraMeshSlicer()
		{
			AddInPort("TetraMesh",typeof(UnstrGrid));
			AddInPort("Plane",typeof(VectorND));
			AddOutPort("TriangleMesh",typeof(TriangleMesh));
		}

        /// <summary>
        /// evaluates a single point in the context of tha slicing plane.
        /// </summary>
        /// <param name="point">A point to be evaluated.</param>
        /// <param name="plane">A plane to be evaluated against (implicit equation coeficients, 4D, i.e. 5 values).</param>
        /// <returns>plane equation value at given point.</returns>
		private double evaluate(PointND point, VectorND plane)
		{
			double result = point.Values[0]*plane.Values[0] +
							point.Values[1]*plane.Values[1] +
							point.Values[2]*plane.Values[2] +
							point.Values[3]*plane.Values[3]
							+ plane.Values[4];
			return(result);
		}

		private int getPointIndex(int a, int b)
		{
			Line l;
			if (a<b)
				l = new Line(a,b);
			else
				l = new Line(b,a);
            int index;
			if (edges.TryGetValue(l,out index))
			{
				return(index);
			}
			else
			{
				PointND p = (PointND)sourcePoints[a];
				PointND v = (PointND)sourcePoints[b]-(PointND)sourcePoints[a];
				double t = -(plane[0]*p[0]+plane[1]*p[1]+plane[2]*p[2]+plane[3]*p[3]+plane[4])/
							(plane[0]*v[0]+plane[1]*v[1]+plane[2]*v[2]+plane[3]*v[3]);
				PointND intersection = p+t*v;
				PointND transformed = intersection*projMatrix;
				Point3D result = new Point3D(0,0,0);
				int j = 0;
				for (int i = 0;i<4;i++)
				{
					if (i!=maxIndex)
					{
						result[j] = transformed[i];
						j++;
					}
				}
				points.Add(result);
				edges.Add(l,(points.Count-1));
				return(points.Count-1);
			}
		}

		private void addTriangle(int e1a, int e1b, int e2a, int e2b, int e3a, int e3b)
		{
			int i1 = getPointIndex(e1a,e1b);
			int i2 = getPointIndex(e2a,e2b);
			int i3 = getPointIndex(e3a,e3b);
			triangles.Add(new Triangle(i1,i2,i3));
		}		

		private Transform4D createProjectionMatrix(VectorND plane)
		{
			maxIndex = 0;
			double max = double.MinValue;
			for (int i = 0;i<4;i++)
			{
				if (plane[i]>max)
				{
					max = plane[i];
					maxIndex = i;
				}
			}

			Transform4D result = new Transform4D();

			for (int i = 0;i<4;i++)
			{
				if (i!= maxIndex)
				{
					Transform4D rot = new Transform4D();
					rot.SetRotation(maxIndex,i,Math.Asin(plane[i]));
					result*=rot;
				}
			}

			Transform4D translation = new Transform4D();
			double nrm = (plane[0]*plane[0]+plane[1]*plane[1]+plane[2]*plane[2]+plane[3]*plane[3]);
			double tx = (plane[0]*plane[4])/nrm;
			double ty = (plane[1]*plane[4])/nrm;
			double tz = (plane[2]*plane[4])/nrm;
			double tw = (plane[3]*plane[4])/nrm;
			translation.SetTranslation(tx,ty,tz,tw);

			result = translation*result;

			return(result);
		}

        /// <summary>
        /// Performs the slicing. 
        /// If the inputs are unchanged and result from previous run is available, then it is set to output and dataSame is set to true.
        /// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(!IsAllDataSame()))
			{
				UnstrGrid input = (UnstrGrid)GetInput("TetraMesh");
				tetrahedra = (UniformDataArray)(input.GetCells(typeof(Tetrahedron))[0]);
				sourcePoints = input.Points;

				plane = (VectorND)GetInput("Plane");
				if (plane.Length!=5)
				{
					throw new MveException(this, "Plane for slicing must be specified by a five component vector.");
				}

				points = new List<Point3D>();
				triangles = new List<Triangle>();
				LineEqualityComparer.pointCount = sourcePoints.Count;
				edges = new Dictionary<Line,int>(tetrahedra.Count*2/3,new LineEqualityComparer());

				projMatrix = createProjectionMatrix(plane);

				evals = new bool[sourcePoints.Count];
				for (int j = 0;j<evals.Length;j++)
				{
					evals[j] = evaluate((PointND)sourcePoints[j],plane)>0;
					if ((j%1000) == 0) ProgressInfo((float)j/evals.Length);
				}

				int tIndex = 0;
				foreach (Tetrahedron t in tetrahedra)
				{
					int index = 0;
					if (evals[t.V1]) index += 1;
					if (evals[t.V2]) index += 2;
					if (evals[t.V3]) index += 4;
					if (evals[t.V4]) index += 8;
					switch (index)
					{
						case 0: // all vertices below
							break;
						case 1: // vertex 1 above, rest below
							addTriangle(t.V1, t.V2, t.V1, t.V3, t.V1, t.V4);
							break;			
						case 2: // vertex 2 above, rest below
							addTriangle(t.V2, t.V1, t.V2, t.V4, t.V2, t.V3);
							break;						
						case 3: // vertices 1 and 2 above, rest below
							addTriangle(t.V1, t.V3, t.V1, t.V4, t.V2, t.V3);
							addTriangle(t.V2, t.V3, t.V1, t.V4, t.V2, t.V4);
							break;
						case 4: // vertex 3 above, rest below
							addTriangle(t.V3, t.V1, t.V3, t.V2, t.V3, t.V4);
							break;
						case 5:
							addTriangle(t.V1, t.V4, t.V1, t.V2, t.V2, t.V3);
							addTriangle(t.V1, t.V4, t.V2, t.V3, t.V3, t.V4);
							break;
						case 6:
							addTriangle(t.V1, t.V3, t.V1, t.V2, t.V3, t.V4);
							addTriangle(t.V2, t.V1, t.V2, t.V4, t.V3, t.V4);
							break;
						case 7:
							addTriangle(t.V4, t.V1, t.V4, t.V2, t.V4, t.V3);
							break;
						case 8:
							addTriangle(t.V4, t.V1, t.V4, t.V3, t.V4, t.V2);
							break;
						case 9:
							addTriangle(t.V1, t.V2, t.V1, t.V3, t.V3, t.V4);
							addTriangle(t.V2, t.V4, t.V2, t.V1, t.V3, t.V4);
							break;
						case 10:
							addTriangle(t.V1, t.V2, t.V1, t.V4, t.V2, t.V3);
							addTriangle(t.V2, t.V3, t.V1, t.V4, t.V3, t.V4);
							break;
						case 11: // vertex 3 below, rest above
							addTriangle(t.V3, t.V2, t.V3, t.V1, t.V3, t.V4);
							break;
						case 12: // vertices 1 and 2 below, rest above
							addTriangle(t.V1, t.V4, t.V1, t.V3, t.V2, t.V3);
							addTriangle(t.V1, t.V4, t.V2, t.V3, t.V2, t.V4);
							break;
						case 13: // vertex 2 below, rest above
							addTriangle(t.V2, t.V4, t.V2, t.V1, t.V2, t.V3);
							break;						
						case 14: // vertex 1 below, rest above
							addTriangle(t.V1, t.V3, t.V1, t.V2, t.V1, t.V4);
							break;			
						case 15:
							break;
					}
					tIndex++;
					if ((tIndex%1000) == 0)
					{
						ProgressInfo((float)tIndex/tetrahedra.Count);
					}
				}

				UniformDataArray trianglesArray = new UniformDataArray(typeof(Triangle),triangles.Count);
				int i = 0;
				foreach(Triangle t in triangles)
				{
					trianglesArray[i] = t;
					i++;
				}

				UniformDataArray pointsArray = new UniformDataArray(typeof(Point3D),points.Count);
				i = 0;
				foreach(Point3D p in points)
				{
					pointsArray[i] = p;
					i++;
				}

				output = new TriangleMesh();
				output.Points = pointsArray;
				output.AddCells(triangleCellsName,trianglesArray);

				SetOutput("TriangleMesh",output,false);
				paramsChanged = false;
			}
			else
			{
				SetOutput("TriangleMesh",output,true);
			}
		}
	}
}
